// ----------------------------------------------------
// Made by: Peepss
// Warning! Do not use this code without my permission.
// ----------------------------------------------------

import config from "../config.js";

import {
    world,
    system,
    ItemStack,
    EnchantmentTypes
} from "@minecraft/server";
import { Enchantments } from "classes/Enchantments.js";

class BackpackEventManager {
    constructor() {
        this.listeners = {
            hold: [],
            release: [],
            inventoryOpen: [],
            inventoryClose: [],
        };
        this.run();
    }

    on(event, callback) {
        if (!this.listeners[event]) {
            throw new Error(`Evento desconhecido: ${event}`);
        }
        if (typeof callback === "function") {
            this.listeners[event].push(callback);
        } else {
            throw new Error("O callback deve ser uma função.");
        }
    }

    emit(event, data) {
        if (!this.listeners[event]) {
            throw new Error(`Evento desconhecido: ${event}`);
        }
        for (const listener of this.listeners[event]) {
            listener(data);
        }
    }

    getAllDynamicProperties(target) {
        const propertyIdentifiers = target.getDynamicPropertyIds();

        if (propertyIdentifiers.length > 0) {
            const dynamicProperties = [];

            propertyIdentifiers.forEach(identifier => {
                const propertyData = target.getDynamicProperty(identifier);

                if (propertyData) {
                    dynamicProperties.push({
                        identifier: identifier,
                        propertyData: propertyData,
                    });
                }
            });

            return dynamicProperties;
        }

        return [];
    }

    saveInventory(player, backpack, item) {
        const inventory = backpack.getComponent("minecraft:inventory").container;
        let isInventoryChanged = false;

        for (let slot = 0; slot < inventory.size; slot++) {
            const itemOnSlot = inventory.getItem(slot);

            let itemData = {
                typeId: "minecraft:air",
                count: 1,
                lore: [],
                enchantments: [],
                durability: null,
                customName: null
            };

            if (itemOnSlot) {
                itemData.typeId = itemOnSlot.typeId;
                itemData.count = itemOnSlot.amount;

                const lore = itemOnSlot.getLore();
                if (lore && lore.length > 0) {
                    itemData.lore = lore;
                }

                const enchantments = Enchantments.getEnchants(itemOnSlot);
                if (enchantments.length > 0) {
                    itemData.enchantments = enchantments.map(enchant => ({
                        name: enchant.type.id,
                        level: enchant.level
                    }));
                }

                const durabilityComponent = itemOnSlot.getComponent("durability");
                if (durabilityComponent) {
                    itemData.durability = durabilityComponent.damage;
                }

                const nameTag = itemOnSlot.nameTag;
                if (nameTag) {
                    itemData.customName = nameTag;
                }
            }

            const currentItemData = JSON.parse(item.getDynamicProperty(`inv_${slot}`) || '{}');
            if (JSON.stringify(itemData) !== JSON.stringify(currentItemData)) {
                item.setDynamicProperty(`inv_${slot}`, JSON.stringify(itemData));
                isInventoryChanged = true;
            }
        }

        if (isInventoryChanged) {
            const playerInventory = player.getComponent("minecraft:inventory").container;

            for (let pSlot = 0; pSlot < playerInventory.size; pSlot++) {
                const playerCurrentSlot = playerInventory.getItem(pSlot);

                if (playerCurrentSlot) {
                    const playerItemData = playerCurrentSlot.getDynamicProperty("itemData");
                    if (
                        playerItemData &&
                        JSON.parse(playerItemData).customId === backpack.getDynamicProperty("customId")
                    ) {
                        playerInventory.setItem(pSlot, item);
                        break;
                    }
                }
            }
        }
    }

    loadInventory(backpack, item) {
        const properties = this.getAllDynamicProperties(item);
        const backpackInventory = backpack.getComponent("minecraft:inventory").container;

        if (!backpackInventory) return;

        properties.forEach(property => {
            try {
                const { identifier, propertyData } = property;
                const slot = parseInt(identifier.replace("inv_", ""));

                //If isn't an inventory slot property, return.
                if (!identifier.includes("inv_")) return;

                const parsedPropertyData = JSON.parse(propertyData);
                const { typeId, count, lore, enchantments, durability, customName } = parsedPropertyData;

                //If typeId equals "minecraft:air", set empty slot.
                if (!typeId || typeId === "minecraft:air") {
                    backpackInventory.setItem(slot, null);
                }

                const itemStack = new ItemStack(typeId, count);

                //Adding lore if the item has lore stored in the itemData property.
                if (lore && lore.length > 0) {
                    if (typeof itemStack.setLore === "function") {
                        itemStack.setLore(lore);
                    }
                }

                //Adding enchantments if the item has any enchantment stored in the itemData property.
                if (enchantments) {
                    for (const enchant of enchantments) {
                        Enchantments.addEnchant(itemStack, { type: EnchantmentTypes.get(enchant.name), level: enchant.level });
                    }
                }

                //Adding the correct durability if the item has any stored in the itemData property.
                if (durability !== null) {
                    const durabilityComponent = itemStack.getComponent("durability");

                    if (durabilityComponent) {
                        durabilityComponent.damage = durability;
                    }
                }

                //Adding the nameTag if the item has any stored in the itemData property.
                if (customName) {
                    itemStack.nameTag = customName;
                }

                //Setting the item
                backpackInventory.setItem(slot, itemStack);
            } catch (error) {
                console.error("Error loading inventory:", error);
            }
        });
    }

    spawnBackpackEntity(player, item) {
        const backpack = player.dimension.spawnEntity("peepss:backpack_container", player.location);
        const customId = JSON.parse(item.getDynamicProperty("itemData")).customId;
        const typeId = JSON.parse(item.getDynamicProperty("itemData")).typeId;

        backpack.setDynamicProperty("owner", player.name);
        backpack.setDynamicProperty("customId", customId);

        const itemName = item.nameTag || "";
        backpack.nameTag = itemName && itemName !== "item." + typeId ? itemName : "item." + typeId;

        const intervalId = system.runInterval(() => {
            if (!backpack.isValid || !player.isValid) {
                backpack.triggerEvent("minecraft:despawn");
                system.clearRun(intervalId);
                return;
            }

            this.saveInventory(player, backpack, item);

            if (backpack.hasTag("inventoryOpen")) {
                backpack.removeTag("inventoryOpen");
                this.emit("inventoryOpen", { player, item });
            }

            if (backpack.hasTag("inventoryClose")) {
                backpack.removeTag("inventoryClose");
                this.emit("inventoryClose", { player, item });
            }

            const playerInventory = player.getComponent("minecraft:inventory");
            const selectedSlotItem = playerInventory?.container?.getItem(player.selectedSlotIndex);

            let itemData = null;
            if (selectedSlotItem && selectedSlotItem.getDynamicProperty("itemData")) {
                itemData = JSON.parse(selectedSlotItem.getDynamicProperty("itemData"));
            }

            if (!itemData || itemData.customId !== customId) {
                backpack.triggerEvent("minecraft:despawn");
                system.clearRun(intervalId);
                return;
            }

            backpack.teleport({
                x: player.location.x,
                y: player.location.y + 1.2,
                z: player.location.z
            }, player.dimension);
        }, 0);

        return backpack;
    }

    run() {
        system.runInterval(() => {
            try {
                for (const player of world.getPlayers()) {
                    const playerInventory = player.getComponent("minecraft:inventory");
                    const selectedSlotItem = playerInventory?.container?.getItem(player.selectedSlotIndex);

                    const lastHeldTypeId = player.getDynamicProperty("lastHeldTypeId");
                    const lastHeldCustomId = player.getDynamicProperty("lastHeldCustomId");

                    if (selectedSlotItem) {
                        const itemData = selectedSlotItem.getDynamicProperty("itemData");

                        if (config.backpacksTypeIds.includes(selectedSlotItem.typeId) && !itemData) {
                            const customId = Math.floor(Math.random() * 9000) + 1000;
                            const newItemData = {
                                typeId: selectedSlotItem.typeId,
                                customId: customId,
                            };

                            const jsonData = JSON.stringify(newItemData);
                            selectedSlotItem.setDynamicProperty("itemData", jsonData);
                            playerInventory.container.setItem(player.selectedSlotIndex, selectedSlotItem);
                        }
                    }

                    if (selectedSlotItem) {
                        const itemData = selectedSlotItem.getDynamicProperty("itemData");
                        let customId = null;

                        if (itemData) {
                            try {
                                const parsedData = JSON.parse(itemData);
                                customId = parsedData?.customId || null;
                            } catch (error) {
                                console.warn(`Erro ao parsear os dados do item: ${error.message}`);
                            }
                        }

                        if (lastHeldTypeId !== selectedSlotItem.typeId || lastHeldCustomId !== customId) {
                            if (lastHeldTypeId && config.backpacksTypeIds.includes(lastHeldTypeId)) {
                                this.emit("release", {
                                    player,
                                    item: {
                                        typeId: lastHeldTypeId,
                                        customId: lastHeldCustomId,
                                    },
                                });
                            }

                            player.setDynamicProperty("lastHeldTypeId", selectedSlotItem.typeId);
                            player.setDynamicProperty("lastHeldCustomId", customId);

                            if (config.backpacksTypeIds.includes(selectedSlotItem.typeId)) {
                                this.emit("hold", { player, item: selectedSlotItem });
                            }
                        }
                    } else if (lastHeldTypeId) {
                        if (config.backpacksTypeIds.includes(lastHeldTypeId)) {
                            this.emit("release", {
                                player,
                                item: {
                                    typeId: lastHeldTypeId,
                                    customId: lastHeldCustomId,
                                },
                            });
                        }

                        player.setDynamicProperty("lastHeldTypeId", null);
                        player.setDynamicProperty("lastHeldCustomId", null);
                    }
                }
            } catch (error) {
                console.error("Erro no loop do Backpack Manager:", error);
            }
        }, 0);
    }
}

const backpackEventManager = new BackpackEventManager();

export default backpackEventManager;